﻿using Infragistics.Controls.Charts;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Media;
using System.Windows.Shapes;
using Frame = Infragistics.Controls.Charts.Frame;
 
namespace Infragistics.Extension
{
    public enum ScatterItemHighlightType
    {
        Auto,
        Marker,
        //Shape, TODO
    }

    public class CustomScatterItemHighlightLayer
        : AnnotationLayer
    {
        /// <summary>
        /// Constructs a CategoryItemHightlightLayer
        /// </summary>
        public CustomScatterItemHighlightLayer()
        {
            var previousFrame = new CustomScatterItemHighlightLayerFrame();
            var currentFrame = new CustomScatterItemHighlightLayerFrame();
            var transitionFrame = new CustomScatterItemHighlightLayerFrame();

            PreviousFrame = previousFrame;
            CurrentFrame = currentFrame;
            TransitionFrame = transitionFrame;

            DefaultStyleKey = typeof (CustomScatterItemHighlightLayer);
        }


        #region Properties
        /// <summary>
        /// The view for this series.
        /// </summary>
        internal CustomScatterItemHighlightLayerView ItemView { get; set; }
      
        #region TargetSeries Dependency Property

        /// <summary>
        /// Gets or sets the series to target this annotation to. If null, this annotation targets all series simultaneously.
        /// <para>
        /// This is a dependency property.
        /// </para>
        /// </summary>
        public Series TargetSeries
        {
            get { return (Series)GetValue(TargetSeriesProperty); }
            set { SetValue(TargetSeriesProperty, value); }
        }

        internal const string TargetSeriesPropertyName = "TargetSeries";

        /// <summary>
        /// Identifies the TargetSeries Dependency Property.
        /// </summary>
        public static readonly DependencyProperty TargetSeriesProperty =
            DependencyProperty.Register(TargetSeriesPropertyName, typeof(Series), typeof(CustomScatterItemHighlightLayer),
                new PropertyMetadata(null, (sender, e) =>
                {
                    ((CustomScatterItemHighlightLayer)sender).RaisePropertyChanged(TargetSeriesPropertyName, e.OldValue,
                        e.NewValue);
                }));

        #endregion

        #region UseInterpolation Dependency Property

        /// <summary>
        /// Gets or sets whether to use value interpolation when drawing a line through the best value for the mouse position.
        /// <para> This is a dependency property.  </para>
        /// </summary>
        public bool UseInterpolation
        {
            get { return (bool)GetValue(UseInterpolationProperty); }
            set { SetValue(UseInterpolationProperty, value); }
        }

        internal const string UseInterpolationPropertyName = "UseInterpolation";

        /// <summary>
        /// Identifies the UseInterpolation Dependency Property.
        /// </summary>
        public static readonly DependencyProperty UseInterpolationProperty = DependencyProperty.Register(
            UseInterpolationPropertyName, typeof(bool), typeof(CustomScatterItemHighlightLayer),
            new PropertyMetadata(false, (sender, e) =>
            {
                ((CustomScatterItemHighlightLayer)sender).RaisePropertyChanged(UseInterpolationPropertyName, e.OldValue,
                    e.NewValue);
            }));

        #endregion

        #region HighlightType Dependency Property

        /// <summary>
        /// Gets or sets which type of highlight shape to use when highlighting items.
        /// <para>
        /// This is a dependency property.
        /// </para>
        /// </summary>
        public ScatterItemHighlightType HighlightType
        {
            get { return (ScatterItemHighlightType)GetValue(HighlightTypeProperty); }
            set { SetValue(HighlightTypeProperty, value); }
        }

        internal const string HighlightTypePropertyName = "HighlightType";

        /// <summary>
        /// Identifies the HighlightType Dependency Property.
        /// </summary>
        public static readonly DependencyProperty HighlightTypeProperty = DependencyProperty.Register(
            HighlightTypePropertyName, typeof(ScatterItemHighlightType), typeof(CustomScatterItemHighlightLayer),
            new PropertyMetadata(ScatterItemHighlightType.Auto, (sender, e) =>
            {
                ((CustomScatterItemHighlightLayer)sender).RaisePropertyChanged(HighlightTypePropertyName, e.OldValue,
                    e.NewValue);
            }));

        #endregion

        #region MarkerType

        /// <summary>
        /// Gets or sets the marker type for the current series object.
        /// </summary>
        /// <remarks>
        /// If the MarkerTemplate property is set, the setting of the MarkerType property will be ignored.
        /// </remarks>
        [WidgetDefaultString("none")]
        public MarkerType MarkerType
        {
            get { return (MarkerType)GetValue(MarkerTypeProperty); }
            set { SetValue(MarkerTypeProperty, value); }
        }

        internal const string MarkerTypePropertyName = "MarkerType";

        /// <summary>
        /// Identifies the MarkerType dependency property.
        /// </summary>
        public static readonly DependencyProperty MarkerTypeProperty = DependencyProperty.Register(
            MarkerTypePropertyName, typeof(MarkerType), typeof(CustomScatterItemHighlightLayer),
            new PropertyMetadata(MarkerType.Circle, (sender, e) =>
            {
                ((CustomScatterItemHighlightLayer)sender).RaisePropertyChanged(MarkerTypePropertyName, e.OldValue, e.NewValue);
            }));

        #endregion

        #region MarkerTemplate Dependency Property

        /// <summary>
        ///  Gets or sets the template to use for marker visuals for the current series object.
        /// </summary>
        public DataTemplate MarkerTemplate
        {
            get { return (DataTemplate)GetValue(MarkerTemplateProperty); }
            set { SetValue(MarkerTemplateProperty, value); }
        }

        internal const string MarkerTemplatePropertyName = "MarkerTemplate";

        /// <summary>
        /// Identifies the MarkerTemplate Dependency Property.
        /// </summary>
        public static readonly DependencyProperty MarkerTemplateProperty =
            DependencyProperty.Register(MarkerTemplatePropertyName, typeof(DataTemplate),
                typeof(CustomScatterItemHighlightLayer),
                new PropertyMetadata(null, (sender, e) =>
                {
                    ((CustomScatterItemHighlightLayer)sender).RaisePropertyChanged(MarkerTemplatePropertyName, e.OldValue,
                        e.NewValue);
                }));

        #endregion

        #region BandHighlightWidth Dependency Property

        /// <summary>
        /// Gets or sets the width to use for the highlight region if highlighting items in a grid aligned series (line, spline, etc), with a banded shape.
        /// <para>
        /// This is a dependency property.
        /// </para>
        /// </summary>
        public double BandHighlightWidth
        {
            get { return (double)GetValue(BandHighlightWidthProperty); }
            set { SetValue(BandHighlightWidthProperty, value); }
        }

        internal const string BandHighlightWidthPropertyName = "BandHighlightWidth";

        /// <summary>
        /// Identifies the ItemHighlightWidth Dependency Property.
        /// </summary>
        public static readonly DependencyProperty BandHighlightWidthProperty =
            DependencyProperty.Register(BandHighlightWidthPropertyName, typeof(double),
                typeof(CustomScatterItemHighlightLayer),
                new PropertyMetadata(10.0, (sender, e) =>
                {
                    ((CustomScatterItemHighlightLayer)sender).RaisePropertyChanged(BandHighlightWidthPropertyName, e.OldValue,
                        e.NewValue);
                }));

        #endregion

        #region SkipUnknownValues Dependency Property

        /// <summary>
        /// Gets or sets whether to skip unknown values when searching for series values.
        /// <para>
        /// This is a dependency property.
        /// </para>
        /// </summary>
        public bool SkipUnknownValues
        {
            get { return (bool)GetValue(SkipUnknownValuesProperty); }
            set { SetValue(SkipUnknownValuesProperty, value); }
        }

        internal const string SkipUnknownValuesPropertyName = "SkipUnknownValues";

        /// <summary>
        /// Identifies the SkipUnknownValues Dependency Property.
        /// </summary>
        public static readonly DependencyProperty SkipUnknownValuesProperty =
            DependencyProperty.Register(SkipUnknownValuesPropertyName, typeof(bool), typeof(CustomScatterItemHighlightLayer),
                new PropertyMetadata(true, (sender, e) =>
                {
                    ((CustomScatterItemHighlightLayer)sender).RaisePropertyChanged(SkipUnknownValuesPropertyName, e.OldValue,
                        e.NewValue);
                }));

        #endregion

        #endregion

        #region Methods Overriden

        /// <summary>
        /// Overridden to provide the view for this series.
        /// </summary>
        /// <returns></returns>
        protected override SeriesView CreateView()
        {
            return new CustomScatterItemHighlightLayerView(this);
        }

        /// <summary>
        /// Overridden to process the created view for this series at creation time.
        /// </summary>
        /// <param name="view">The created view.</param>
        protected override void OnViewCreated(SeriesView view)
        {
            base.OnViewCreated(view);

            ItemView = (CustomScatterItemHighlightLayerView)view;
        }

        /// <summary>
        /// Overridden to react to when a series on which the annotation layer is possibly dependent is rendering.
        /// </summary>
        /// <param name="series">The series being rendered.</param>
        /// <param name="animate">Whether animation was requested when the series was rendered.</param>
        protected override void OnDependentSeriesRender(Series series, bool animate)
        {
            if (TargetSeries == null || TargetSeries == series)
            {
                //if (!ContentInfo.IsDirty)
                //{
                RenderSeries(animate);
                //}
            }
        }

        protected override bool IsSeriesValid(Series series)
        {
            if (!base.IsSeriesValid(series))
            {
                return false;
            }
            var res = GetCategoryPosition(series);
            if (!res.Item1)
            {
                return false;
            }
            return true;
        }

        /// <summary>
        /// Called when the series should prepare the next animation frame.
        /// </summary>
        /// <param name="frame">The animation frame to populate.</param>
        /// <param name="view">The view for which we are preparing a frame.</param>
        protected override void PrepareFrame(Frame frame, SeriesView view)
        {
            base.PrepareFrame(frame, view);

            var f = (CustomScatterItemHighlightLayerFrame) frame;
            var useInterpolation = this.UseInterpolation;
            f.Rects.Clear();

            if (double.IsNaN(CurrentPosition.X) &&
                double.IsNaN(CurrentPosition.Y))
            {
                return;
            }

            if (IsSeriesValid(TargetSeries))
            {
                PrepareSeries(TargetSeries, f, useInterpolation);
            }
            else
            {
                foreach (var series in SeriesViewer.Series)
                {
                    if (IsSeriesValid(series))
                    {
                        PrepareSeries(series, f, useInterpolation);
                    }
                }
            }
        }

        /// <summary>
        /// Called when the series should render the provided animation frame.
        /// </summary>
        /// <param name="frame">The frame to be rendered.</param>
        /// <param name="view">The view to be rendered.</param>
        protected override void RenderFrame(Frame frame, SeriesView view)
        {
            base.RenderFrame(frame, view);

            var rectPool = ItemView.RectPool;
            var markerPool = ItemView.MarkerPool;

            var rectCount = 0;
            var markerCount = 0;

            var f = (CustomScatterItemHighlightLayerFrame) frame;

            foreach (var rect in f.Rects.Values)
            {
                if (rect.MarkerTemplate != null)
                {
                    var markerVisual = markerPool[markerCount];
                    RenderPoint(rect, markerVisual);
                    markerCount++;
                }
                else
                {
                    var rectVisual = rectPool[rectCount];
                    RenderRect(rect, rectVisual);
                    rectCount++;
                }
            }
            rectPool.Count = rectCount;
            markerPool.Count = markerCount;

            //System.Diagnostics.Debug.WriteLine("lineCount: " + lineCount);
        }

        /// <summary>
        /// Overridden to react to changes in series property values.
        /// </summary>
        /// <param name="sender">The series being altered.</param>
        /// <param name="propertyName">The property name of the property being altered.</param>
        /// <param name="oldValue">The old value of the property.</param>
        /// <param name="newValue">THe new value of the property.</param>
        protected override void PropertyUpdatedOverride(object sender, string propertyName, object oldValue,
            object newValue)
        {
            base.PropertyUpdatedOverride(sender, propertyName, oldValue, newValue);

            switch (propertyName)
            {
                case TargetSeriesPropertyName:
                    RenderSeries(true);
                    break;
                case UseInterpolationPropertyName:
                    RenderSeries(true);
                    break;
                case HighlightTypePropertyName:
                    RenderSeries(true);
                    break;
                case BandHighlightWidthPropertyName:
                    RenderSeries(true);
                    break;
                case SkipUnknownValuesPropertyName:
                    RenderSeries(true);
                    break;
            }
        }

        #endregion

        #region Methods Private

        private CustomDataContext GetDataItem(Series target, Point pt)
        {
            var item = target.GetItem(pt);
            var exactIndex = target.GetExactItemIndex(pt);

            if (UseInterpolation &&
                exactIndex != Math.Floor(exactIndex))
            {
                item = null;
            }

            return new CustomDataContext()
            {
                Series = target,
                Item = item,
                ItemBrush = new SolidColorBrush(Colors.Red)
            };
        }
        
        private void PrepareSeries(Series series, CustomScatterItemHighlightLayerFrame f, bool useInterpolation)
        {
            if (this.View == null ||
                this.SeriesViewer == null ||
                this.SeriesViewer.ActualWindowRect.IsEmpty ||
                this.Viewport.IsEmpty)
            {
                return;
            }

            if (!series.IsScatter)
            {
                return;
            }

            var skipUnknowns = SkipUnknownValues;

            bool usePointType;

            if (this.HighlightType == ScatterItemHighlightType.Auto ||
                this.HighlightType == ScatterItemHighlightType.Marker)
            {
                usePointType = true;
            }
            else
            {
                usePointType = false;
            }

            var res = GetCategoryPosition(series);
            if (!res.Item1) // skip if cannot plot item
            {
                return;
            }
            if (usePointType)
            {
                //Debug.WriteLine("PrepareSeries");
                var pt = series.GetSeriesValuePosition(res.Item2, useInterpolation, skipUnknowns);
                //var pt = res.Item2;
                //if (pt.IsPlottable())
                //{
                //    pt = ToWorldPosition(pt);
                //}
                PreparePoint(pt, f, series);
            }
            else
            {
                //TODO plot other type of highlighting for scatter item
                //var pt = series.GetSeriesValuePosition(res.Item2, UseInterpolation, skipUnknowns);
                var rect = Rect.Empty;

                #region Implementation

                //if (series.IsVertical)
                //{
                //    var y = pt.Y;
                //    var catAxis = cat.CategoryAxis;
                //    double groupWidth = _mode0HighlightWidth;

                //    if (preferredMode != CategoryMode.Mode0)
                //    {
                //        groupWidth = catAxis.GetGroupSize(GetWindowRect(), this.Viewport);
                //    }
                //    if (double.IsNaN(groupWidth) || double.IsInfinity(groupWidth))
                //    {
                //        return;
                //    }
                //    rect = new Rect(Viewport.Left, y - groupWidth / 2.0, Viewport.Width, groupWidth);
                //}
                //else
                //{
                //    var x = pt.X;
                //    var catAxis = cat.CategoryAxis;
                //    double groupWidth = _mode0HighlightWidth;

                //    if (preferredMode != CategoryMode.Mode0)
                //    {
                //        groupWidth = catAxis.GetGroupSize(GetWindowRect(), this.Viewport);
                //    }
                //    if (double.IsNaN(groupWidth) || double.IsInfinity(groupWidth))
                //    {
                //        return;
                //    }
                //    rect = new Rect(x - groupWidth / 2.0, Viewport.Top, groupWidth, Viewport.Height);
                //} 

                #endregion

                rect.Intersect(this.Viewport);

                PrepareRect(rect, f, series);
            }
        }

        private void PreparePoint(Point pt, CustomScatterItemHighlightLayerFrame frame, Series target)
        {
            var n = new ItemHighlightFrameRect();

            n.Left = pt.X;
            n.Top = pt.Y;

            n.Brush = GetBrush();
            n.Outline = GetOutline();

            if (n.Brush == null)
            {
                if (target.HasMarkers)
                {
                    n.Brush = ItemView.GetLightenedBrush(target.GetActualMarkerBrush());
                }
                else
                {
                    n.Brush = ItemView.GetLightenedBrush(target.ActualBrush);
                }
            }
            if (n.Outline == null)
            {
                if (target.HasMarkers)
                {
                    n.Outline = ItemView.GetLightenedBrush(target.GetActualMarkerOutlineBrush());
                }
                else
                {
                    n.Outline = ItemView.GetLightenedBrush(target.ActualOutline);
                }
            }

            n.StrokeThickness = Thickness;
            n.TimeStamp = target.GetHashCode();
            n.DataItem = GetDataItem(target, pt);

            #region targetTemplate

            DataTemplate targetTemplate = null; //SeriesViewer.CircleMarkerTemplate;

            //var targetTemplate = (target is MarkerSeries) ? ((MarkerSeries)target).ActualMarkerTemplate : null;
            //var markerType = MarkerSeries.ResolveMarkerType(target, MarkerType.Automatic);

            switch (this.MarkerType)
            {
                case MarkerType.Circle:
                    targetTemplate = SeriesViewer.CircleMarkerTemplate;
                    break;
                case MarkerType.Triangle:
                    targetTemplate = SeriesViewer.TriangleMarkerTemplate;
                    break;
                case MarkerType.Pyramid:
                    targetTemplate = SeriesViewer.PyramidMarkerTemplate;
                    break;
                case MarkerType.Square:
                    targetTemplate = SeriesViewer.SquareMarkerTemplate;
                    break;
                case MarkerType.Diamond:
                    targetTemplate = SeriesViewer.DiamondMarkerTemplate;
                    break;
                case MarkerType.Pentagon:
                    targetTemplate = SeriesViewer.PentagonMarkerTemplate;
                    break;
                case MarkerType.Hexagon:
                    targetTemplate = SeriesViewer.HexagonMarkerTemplate;
                    break;
                case MarkerType.Tetragram:
                    targetTemplate = SeriesViewer.TetragramMarkerTemplate;
                    break;
                case MarkerType.Pentagram:
                    targetTemplate = SeriesViewer.PentagramMarkerTemplate;
                    break;
                case MarkerType.Hexagram:
                    targetTemplate = SeriesViewer.HexagramMarkerTemplate;
                    break;
            }

            #endregion

            n.MarkerTemplate = this.MarkerTemplate != null ? this.MarkerTemplate : targetTemplate;

            if (!double.IsNaN(n.Left) &&
                !double.IsNaN(n.Top))
            {
                frame.Rects.Add(n.TimeStamp, n);
            }
        }

        private void PrepareRect(Rect rect, CustomScatterItemHighlightLayerFrame frame, Series target)
        {
            ItemHighlightFrameRect n = new ItemHighlightFrameRect();
            n.Left = rect.Left;
            n.Top = rect.Top;
            n.Right = rect.Right;
            n.Bottom = rect.Bottom;

            n.Brush = GetBrush();
            n.Outline = GetOutline();
            if (n.Brush == null)
            {
                if (target != null)
                {
                    n.Brush = ItemView.GetLightenedTranslucentBrush(target.ActualBrush);
                }
            }
            if (n.Outline == null)
            {
                if (target != null)
                {
                    n.Outline = ItemView.GetLightenedTranslucentBrush(target.ActualOutline);
                }
            }

            n.StrokeThickness = Thickness;

            n.TimeStamp = target.GetHashCode();

            if (!rect.IsEmpty)
            {
                frame.Rects.Add(n.TimeStamp, n);
            }
        }

        private void RenderPoint(ItemHighlightFrameRect rect, ContentControl markerVisual)
        {
            if (double.IsNaN(rect.Left) ||
                double.IsNaN(rect.Top) ||
                double.IsNaN(rect.StrokeThickness) ||
                rect.DataItem == null)
            {
                markerVisual.Visibility = Visibility.Collapsed;
                return;
            }

            var dc = (CustomDataContext)rect.DataItem;
            dc.ItemBrush = rect.Brush;

            //var series = ((Series) dc.Series);
            //var pt = new Point(rect.Left, rect.Top);
            //var markerPoint = series.GetSeriesValuePosition(pt, _useInterpolation, _skipUnknownValues);
            var markerPoint = new Point(rect.Left, rect.Top);

            if (!markerPoint.IsPlottable())
            {
                markerVisual.Visibility = Visibility.Collapsed;
                return;
            }

            ItemView.PositionMarker(markerVisual, markerPoint.X, markerPoint.Y);
            markerVisual.ContentTemplate = rect.MarkerTemplate;
            markerVisual.Content = rect.DataItem;

        }

        private void RenderRect(ItemHighlightFrameRect rect, Rectangle rectVisual)
        {
            if (double.IsNaN(rect.Left) ||
                double.IsNaN(rect.Top) ||
                double.IsNaN(rect.Bottom) ||
                double.IsNaN(rect.Right) ||
                double.IsNaN(rect.StrokeThickness))
            {
                rectVisual.Visibility = Visibility.Collapsed;
                return;
            }
            ItemView.PositionRectangle(rectVisual, rect.Left, rect.Top, rect.Right - rect.Left, rect.Bottom - rect.Top);
            rectVisual.Fill = rect.Brush;
            rectVisual.Stroke = rect.Outline;
            rectVisual.StrokeThickness = rect.StrokeThickness;

            if (DashArray != null)
            {
#if TINYCLR
                rectVisual.StrokeDashArray = DashArray;
#else
                rectVisual.StrokeDashArray = DashArray.Duplicate();
#endif
            }
            else
            {
                rectVisual.StrokeDashArray = null;
            }
#if !TINYCLR
            rectVisual.StrokeDashCap = DashCap;
            rectVisual.StrokeLineJoin = LineJoin;
            rectVisual.StrokeStartLineCap = StartCap;
            rectVisual.StrokeEndLineCap = EndCap;
            rectVisual.StrokeMiterLimit = MiterLimit;
#else
            rectVisual.StrokeDashCap = (int)DashCap;
#endif
        }

        #endregion
    
    }
 
     /// <summary>
     /// Represents an animation frame for the CustomScatterItemHighlightLayer
     /// </summary>
     [EditorBrowsable(EditorBrowsableState.Never)]
     public class CustomScatterItemHighlightLayerFrame
         : Frame
     {
         /// <summary>
         /// Constructs a frame for the CategoryHightlightLayer
         /// </summary>
         public CustomScatterItemHighlightLayerFrame()
         {

             Rects = new Dictionary<long, ItemHighlightFrameRect>();

             _interpolator = new DictInterpolator<long, ItemHighlightFrameRect, CustomScatterItemHighlightLayerFrame>(
                 InterpolateRect,
                 (r) => r.TimeStamp,
                 (r) => true,
                 () => new ItemHighlightFrameRect());
 
         }

         /// <summary>
         /// The rectangles to be displayed for this animation frame.
         /// </summary>
         public Dictionary<long, ItemHighlightFrameRect> Rects { get; set; }

         private DictInterpolator<long, ItemHighlightFrameRect, CustomScatterItemHighlightLayerFrame> _interpolator;

         /// <summary>
         /// Called to interpolate a rectangle for an animation frame.
         /// </summary>
         /// <param name="interpolated">The interpolated result rectangle</param>
         /// <param name="p">The interpolation value.</param>
         /// <param name="min">The minimum rectangle.</param>
         /// <param name="max">The maximum rectangle.</param>
         /// <param name="minFrame">The minimum frame.</param>
         /// <param name="maxFrame">The maximum frame.</param>
         protected void InterpolateRect(
             ItemHighlightFrameRect interpolated, float p,
             ItemHighlightFrameRect min, ItemHighlightFrameRect max,
             CustomScatterItemHighlightLayerFrame minFrame, CustomScatterItemHighlightLayerFrame maxFrame)
         {
             if (max != null)
             {
                 interpolated.TimeStamp = max.TimeStamp;
             }
             else if (min != null)
             {
                 interpolated.TimeStamp = min.TimeStamp;
             }
             else
             {
                 interpolated.TimeStamp = 0;
             }
             if (min == null || double.IsNaN(min.Top))
             {
                 interpolated.Top = max != null ? max.Top : double.NaN;
             }
             else if (max == null || double.IsNaN(max.Top))
             {
                 interpolated.Top = double.NaN;
             }
             else
             {
                 interpolated.Top = min.Top + p * (max.Top - min.Top);
             }
             if (min == null || double.IsNaN(min.Left))
             {
                 interpolated.Left = max != null ? max.Left : double.NaN;
             }
             else if (max == null || double.IsNaN(max.Left))
             {
                 interpolated.Left = double.NaN;
             }
             else
             {
                 interpolated.Left = min.Left + p * (max.Left - min.Left);
             }
             if (min == null || double.IsNaN(min.Right))
             {
                 interpolated.Right = max != null ? max.Right : double.NaN;
             }
             else if (max == null || double.IsNaN(max.Right))
             {
                 interpolated.Right = double.NaN;
             }
             else
             {
                 interpolated.Right = min.Right + p * (max.Right - min.Right);
             }
             if (min == null || double.IsNaN(min.Bottom))
             {
                 interpolated.Bottom = max != null ? max.Bottom : double.NaN;
             }
             else if (max == null || double.IsNaN(max.Bottom))
             {
                 interpolated.Bottom = double.NaN;
             }
             else
             {
                 interpolated.Bottom = min.Bottom + p * (max.Bottom - min.Bottom);
             }

             if (min == null || double.IsNaN(min.StrokeThickness))
             {
                 interpolated.StrokeThickness = max != null ? max.StrokeThickness : double.NaN;
             }
             else if (max == null || double.IsNaN(max.StrokeThickness))
             {
                 interpolated.StrokeThickness = double.NaN;
             }
             else
             {
                 interpolated.StrokeThickness = min.StrokeThickness + p * (max.StrokeThickness - min.StrokeThickness);
             }

             if (min == null || min.Brush == null)
             {
                 interpolated.Brush = max != null ? max.Brush : null;
             }
             else if (max == null || max.Brush == null)
             {
                 interpolated.Brush = null;
             }
             else
             {
                 interpolated.Brush = BrushUtil.GetInterpolation(min.Brush, p, max.Brush, InterpolationMode.RGB);
             }
             if (min == null || min.Outline == null)
             {
                 interpolated.Outline = max != null ? max.Outline : null;
             }
             else if (max == null || max.Outline == null)
             {
                 interpolated.Outline = null;
             }
             else
             {
                 interpolated.Outline = BrushUtil.GetInterpolation(min.Outline, p, max.Outline, InterpolationMode.RGB);
             }

             if (min == null || min.MarkerTemplate == null)
             {
                 interpolated.MarkerTemplate = max != null ? max.MarkerTemplate : null;
             }
             else if (max == null || max.Outline == null)
             {
                 interpolated.MarkerTemplate = null;
             }
             else
             {
                 interpolated.MarkerTemplate = max.MarkerTemplate;
             }
             if (min == null || min.DataItem == null)
             {
                 interpolated.DataItem = max != null ? max.DataItem : null;
             }
             else if (max == null || max.Outline == null)
             {
                 interpolated.DataItem = null;
             }
             else
             {
                 interpolated.DataItem = max.DataItem;
             }
         }

         /// <summary>
         /// Called when the frame must be interpolated.
         /// </summary>
         /// <param name="p">The current interpolation value.</param>
         /// <param name="min">The minimum frame to interpolate from.</param>
         /// <param name="max">The maximum frame to interpolate to.</param>
         public override void Interpolate(float p, Frame min, Frame max)
         {
             var minFrame = (CustomScatterItemHighlightLayerFrame)min;
             var maxFrame = (CustomScatterItemHighlightLayerFrame)max;

             _interpolator.Interpolate(Rects, p, minFrame.Rects, maxFrame.Rects, minFrame, maxFrame);
         }
     }


    /// <summary>
    /// Represents a view for the CategoryItemHightlightLayer
    /// </summary>
    [EditorBrowsable(EditorBrowsableState.Never)]
     internal class CustomScatterItemHighlightLayerView
        : AnnotationLayerView
    {
        /// <summary>
        /// Constructs a CategoryItemHighlightView
        /// </summary>
        /// <param name="model"></param>
        public CustomScatterItemHighlightLayerView(CustomScatterItemHighlightLayer model)
            : base(model)
        {
            ItemModel = model;
        }

        /// <summary>
        /// The model for this view.
        /// </summary>
        protected CustomScatterItemHighlightLayer ItemModel { get; set; }

        private Canvas _rectCanvas = new Canvas() { UseLayoutRounding = false };
        private Canvas _markerCanvas = new Canvas();

        /// <summary>
        /// Gets or sets the pool to use to obtain rectangles for display.
        /// </summary>
        protected internal Pool<Rectangle> RectPool { get; set; }
        /// <summary>
        /// Gets or sets the pool to use to obtain markers for display.
        /// </summary>
        protected internal Pool<ContentControl> MarkerPool { get; set; }

        /// <summary>
        /// Called when this view is initialized.
        /// </summary>
        public override void OnInit()
        {
            base.OnInit();

            RectPool = new Pool<Rectangle>()
            {
                Create = RectCreate,
                Activate = RectActivate,
                Disactivate = RectDisactivate,
                Destroy = RectDestroy
            };

            MarkerPool = new Pool<ContentControl>()
            {
                Create = MarkerCreate,
                Activate = MarkerActivate,
                Disactivate = MarkerDisactivate,
                Destroy = MarkerDestroy
            };
        }

        /// <summary>
        /// Called when the template is provided.
        /// </summary>
        protected override void OnTemplateProvided()
        {
            base.OnTemplateProvided();

            if (_rectCanvas != null)
            {
                _rectCanvas.Detach();
            }
            if (_markerCanvas != null)
            {
                _markerCanvas.Detach();
            }
            RootCanvas.Children.Add(_rectCanvas);
            RootCanvas.Children.Add(_markerCanvas);
        }

        private Rectangle RectCreate()
        {
            var rect = new Rectangle();
            rect.IsHitTestVisible = false;
            _rectCanvas.Children.Add(rect);

            return rect;
        }

        private void RectActivate(Rectangle rect)
        {
            rect.Visibility = Visibility.Visible;
        }

        private void RectDisactivate(Rectangle rect)
        {
            rect.Visibility = Visibility.Collapsed;
        }

        private void RectDestroy(Rectangle rect)
        {
            _rectCanvas.Children.Remove(rect);
        }

        private ContentControl MarkerCreate()
        {
            var marker = new ContentControl();
            marker.IsHitTestVisible = false;
            _markerCanvas.Children.Add(marker);

            return marker;
        }

        private void MarkerActivate(ContentControl marker)
        {
            marker.Visibility = Visibility.Visible;
        }

        private void MarkerDisactivate(ContentControl marker)
        {
            marker.Visibility = Visibility.Collapsed;
        }

        private void MarkerDestroy(ContentControl marker)
        {
            _markerCanvas.Children.Remove(marker);
        }

        /// <summary>
        /// Gets a lightened brush from a base brush.
        /// </summary>
        /// <param name="brush">The brush to lighten.</param>
        /// <returns>The lightened brush.</returns>
        public Brush GetLightenedBrush(Brush brush)
        {
            if (brush != null)
            {
                var b = brush.GetLightened(.3);
                return b;
            }
            else
            {
                return null;
            }
        }

        /// <summary>
        /// Gets a lightened and translucent brush from a base brush.
        /// </summary>
        /// <param name="brush">The brush to lighten.</param>
        /// <returns>The lightened and translucent brush.</returns>
        public Brush GetLightenedTranslucentBrush(Brush brush)
        {
            if (brush != null)
            {
                var b = brush.GetLightened(.3);
                b.Opacity = .3;
                return b;
            }
            else
            {
                return null;
            }
        }

        internal void PositionRectangle(Rectangle rectVisual, double left, double top, double width, double height)
        {
            rectVisual.Visibility = Visibility.Visible;
            Canvas.SetLeft(rectVisual, left);
            Canvas.SetTop(rectVisual, top);
            rectVisual.Width = width;
            rectVisual.Height = height;
        }

        internal void PositionMarker(ContentControl markerVisual, double x, double y)
        {
            markerVisual.Measure(new Size(double.PositiveInfinity, double.PositiveInfinity));
            var xOffset = 0 - 0.5 * markerVisual.DesiredSize.Width;
            var yOffset = 0 - 0.5 * markerVisual.DesiredSize.Height;

            Canvas.SetLeft(markerVisual, x + xOffset);
            Canvas.SetTop(markerVisual, y + yOffset);
        }
        ///// <summary>
        ///// Exports visual data for shapes of the layer
        ///// </summary>
        //protected internal override void ExportViewShapes(VisualData.SeriesVisualData svd)
        //{
        //    base.ExportViewShapes(svd);

        //    for (var i = 0; i < _rectCanvas.Children.Count; i++)
        //    {
        //        if (_rectCanvas.Children[i].Visibility == Visibility.Visible)
        //        {
        //            var lineData = new RectangleVisualData("cihRect", (Rectangle)_rectCanvas.Children[i]);
        //            lineData.Tags.Add("Main");
        //            lineData.Tags.Add("categoryItemHighlight");
        //            svd.Shapes.Add(lineData);
        //        }
        //    }

        //    for (var i = 0; i < _markerCanvas.Children.Count; i++)
        //    {
        //        if (_markerCanvas.Children[i].Visibility == Visibility.Visible)
        //        {
        //            svd.MarkerShapes.Add(DoToAllMarkers(_markerCanvas.Children[i] as ContentControl));
        //        }
        //    }
        //}

//        private MarkerVisualData DoToAllMarkers(ContentControl m)
//        {
//            MarkerVisualData mvd = new MarkerVisualData();
//            PrimitiveAppearanceData appearance = new PrimitiveAppearanceData();

//            //note TINYCLR centers around the top left on the marker, Xaml the top left has been adjusted so that the center
//            //is in the appropriate place, so if we are exporting the x y of the marker, export the center in both cases.
//#if TINYCLR
//                    mvd.X = m.CanvasLeft;
//                    mvd.Y = m.CanvasTop;
//#else
//            var tt = m.RenderTransform as TranslateTransform;
//            if (tt != null)
//            {
//                //TODO: why doesn't ActualSize work here? Might actually be an issue with the marker placement.
//                mvd.X = tt.X + m.DesiredSize.Width / 2.0;
//                mvd.Y = tt.Y + m.DesiredSize.Height / 2.0;
//            }
//#endif
//            mvd.Index = -1;
//            mvd.ContentTemplate = m.ContentTemplate;
//            if (m.Content != null && m.Content is CustomDataContext && m.Visibility == Visibility.Visible)
//            {
//                var CustomDataContext = (CustomDataContext)m.Content;
//#if OBJC
//                        mvd.CustomMarkerView = CustomDataContext.CustomMarkerView;
//                        appearance.Fill = CustomDataContext.ActualItemBrush.Color;
//                        appearance.Stroke = CustomDataContext.Outline.Color;
//                        appearance.StrokeThickness = CustomDataContext.Thickness;
//#else
//                appearance.Fill = AppearanceHelper.FromBrush(CustomDataContext.ActualItemBrush);
//#if TINYCLR
//                        appearance.Stroke = AppearanceHelper.FromBrush(CustomDataContext.Outline);
//#else
//                appearance.Stroke = AppearanceHelper.FromBrush(CustomDataContext.ActualItemOutline);
//#endif
//                //appearance.StrokeThickness = this.Thickness;
//#endif

//                //if (CustomDataContext.Item != null)
//                //{
//                //    mvd.Index = FastItemsSource.IndexOf(CustomDataContext.Item);
//                //}
//            }

//            mvd.Visibility = m.Visibility;
//            mvd.MarkerAppearance = appearance;

//            if (m.ContentTemplate == this.Model.SeriesViewer.CircleMarkerTemplate)
//            {
//                mvd.MarkerType = "Circle";
//            }
//            else if (m.ContentTemplate == this.Model.SeriesViewer.DiamondMarkerTemplate)
//            {
//                mvd.MarkerType = "Diamond";
//            }
//            else if (m.ContentTemplate == this.Model.SeriesViewer.HexagonMarkerTemplate)
//            {
//                mvd.MarkerType = "Hexagon";
//            }
//            else if (m.ContentTemplate == this.Model.SeriesViewer.HexagramMarkerTemplate)
//            {
//                mvd.MarkerType = "Hexagram";
//            }
//            else if (m.ContentTemplate == this.Model.SeriesViewer.PentagonMarkerTemplate)
//            {
//                mvd.MarkerType = "Pentagon";
//            }
//            else if (m.ContentTemplate == this.Model.SeriesViewer.PentagramMarkerTemplate)
//            {
//                mvd.MarkerType = "Pentagram";
//            }
//            else if (m.ContentTemplate == this.Model.SeriesViewer.PyramidMarkerTemplate)
//            {
//                mvd.MarkerType = "Pyramid";
//            }
//            else if (m.ContentTemplate == this.Model.SeriesViewer.SquareMarkerTemplate)
//            {
//                mvd.MarkerType = "Square";
//            }
//            else if (m.ContentTemplate == this.Model.SeriesViewer.TetragramMarkerTemplate)
//            {
//                mvd.MarkerType = "Tetragram";
//            }
//            else if (m.ContentTemplate == this.Model.SeriesViewer.TriangleMarkerTemplate)
//            {
//                mvd.MarkerType = "Triangle";
//            }
//            else
//            {
//                mvd.MarkerType = "None";
//            }

//            return mvd;
//        }
   
    }
    
    internal class DictInterpolator<TKey, TValue, TFrame> where TValue : new()
    {
        public DictInterpolator(
#if MOBILE || WINDOWS_PHONE || MONO
        InterpolatePoint interpolatePointStrat,
#else
Action<TValue, float, TValue, TValue, TFrame, TFrame> interpolatePointStrat,
#endif
 Func<TValue, TKey> getKeyStrat,
        Func<TValue, bool> validPointStrat,
        Func<TValue> createValueStrat
        )
        {
            InterpolatePointStrat = interpolatePointStrat;
            GetKeyStrat = getKeyStrat;
            ValidPointStrat = validPointStrat;
            CreateValueStrat = createValueStrat;
        }
#if MOBILE || WINDOWS_PHONE || MONO
        internal delegate void InterpolatePoint(TValue a, float b, TValue c, TValue d, TFrame e, TFrame f);
        private InterpolatePoint InterpolatePointStrat { get; set; }
#else
        private Action<TValue, float, TValue, TValue, TFrame, TFrame> InterpolatePointStrat { get; set; }
#endif
        private Func<TValue, TKey> GetKeyStrat { get; set; }
        private Func<TValue, bool> ValidPointStrat { get; set; }
        private Func<TValue> CreateValueStrat { get; set; }

        public void Interpolate(
            IDictionary<TKey, TValue> target,
            float p,
            IDictionary<TKey, TValue> min,
            IDictionary<TKey, TValue> max,
            TFrame minFrame,
            TFrame maxFrame
            )
        {
            List<TKey> removeFromTarget = new List<TKey>();
            foreach (var key in target.Keys)
            {
                TValue minValue;
                TValue maxValue;
                TValue targetValue = target[key];
                bool minContains = min.TryGetValue(key, out minValue);
                bool maxContains = max.TryGetValue(key, out maxValue);

                if (!minContains &&
                    !maxContains)
                {
                    removeFromTarget.Add(key);
                }
                else
                {
                    InterpolatePointStrat(targetValue, p, minValue, maxValue, minFrame, maxFrame);
                    if (!ValidPointStrat(targetValue))
                    {
                        removeFromTarget.Add(key);
                    }
                }
            }
            foreach (var key in removeFromTarget)
            {
                target.Remove(key);
            }

            foreach (var key in min.Keys)
            {
                TValue minValue = min[key];
                TValue maxValue;
                TValue targetValue;
                max.TryGetValue(key, out maxValue);
                bool targetContains = target.TryGetValue(key, out targetValue);

                if (!targetContains)
                {
                    targetValue = CreateValueStrat();
                    InterpolatePointStrat(targetValue, p, minValue, maxValue, minFrame, maxFrame);
                    if (!ValidPointStrat(targetValue))
                    {
                        continue;
                    }
                    target.Add(GetKeyStrat(targetValue), targetValue);
                }
            }

            foreach (var key in max.Keys)
            {
                TValue maxValue = max[key];
                TValue minValue;
                TValue targetValue;
                bool minContains = min.TryGetValue(key, out minValue);
                bool targetContains = target.TryGetValue(key, out targetValue);

                if (!targetContains && !minContains)
                {
                    targetValue = CreateValueStrat();
                    InterpolatePointStrat(targetValue, p, minValue, maxValue, minFrame, maxFrame);
                    if (!ValidPointStrat(targetValue))
                    {
                        continue;
                    }
                    target.Add(GetKeyStrat(targetValue), targetValue);
                }
            }
        }
    }


    /// <summary>
    /// Represents a CustomDataContext for tooltips, markerItems, cursors and legend items
    /// </summary>
    [EditorBrowsable(EditorBrowsableState.Never)]
    internal class CustomDataContext : DependencyObject, INotifyPropertyChanged
    {
        /// <summary>
        /// Identifies the Series dependency property.
        /// </summary>
        public static readonly DependencyProperty SeriesProperty =
            DependencyProperty.Register(
            "Series",
            typeof(Control),
            typeof(CustomDataContext),
            new PropertyMetadata(null,
                (o, e) => (o as CustomDataContext).RaisePropertyChanged("Series", e.OldValue, e.NewValue)));

        /// <summary>
        /// Gets the current DataContext's series.
        /// </summary>
        public Control Series
        {
            get { return (Control)GetValue(SeriesProperty); }
            set { SetValue(SeriesProperty, value); }
        }

        /// <summary>
        /// Identifies the Item dependency property.
        /// </summary>
        public static readonly DependencyProperty ItemProperty =
            DependencyProperty.Register(
            "Item",
            typeof(object),
            typeof(CustomDataContext),
            new PropertyMetadata(null,
                (o, e) => (o as CustomDataContext).RaisePropertyChanged("Item", e.OldValue, e.NewValue)));

        /// <summary>
        /// Gets the current DataContext's item.
        /// </summary>
        public object Item
        {
            get { return GetValue(ItemProperty); }
            set { SetValue(ItemProperty, value); }
        }

        /// <summary>
        /// Identifies the ItemBrush dependency property.
        /// </summary>
        public static readonly DependencyProperty ItemBrushProperty =
            DependencyProperty.Register(
            "ItemBrush",
            typeof(Brush),
            typeof(CustomDataContext),
            new PropertyMetadata(null,
                (o, e) => (o as CustomDataContext).RaisePropertyChanged("ItemBrush", e.OldValue, e.NewValue)));

        /// <summary>
        /// Gets or sets the brush to use to color the outline of the item.
        /// </summary>
        public Brush ItemBrush
        {
            get
            {
                return (Brush)GetValue(ItemBrushProperty);
            }
            set
            {
                SetValue(ItemBrushProperty, value);
            }
        }

        /// <summary>
        /// Identifies the ItemOutline dependency property.
        /// </summary>
        public static readonly DependencyProperty ItemOutlineProperty =
            DependencyProperty.Register(
            "ItemOutline",
            typeof(Brush),
            typeof(CustomDataContext),
            new PropertyMetadata(null,
                (o, e) => (o as CustomDataContext).RaisePropertyChanged("ItemOutline", e.OldValue, e.NewValue)));

        /// <summary>
        /// Gets or sets the brush to use to color the item.
        /// </summary>
        public Brush ItemOutline
        {
            get
            {
                return (Brush)GetValue(ItemOutlineProperty);
            }
            internal set
            {
                SetValue(ItemOutlineProperty, value);
            }
        }

        /// <summary>
        /// Identifies the ActualItemBrush dependency property.
        /// </summary>
        public static readonly DependencyProperty ActualItemBrushProperty =
            DependencyProperty.Register(
            "ActualItemBrush",
            typeof(Brush),
            typeof(CustomDataContext),
            new PropertyMetadata(null,
                (o, e) => (o as CustomDataContext).RaisePropertyChanged("ActualItemBrush", e.OldValue, e.NewValue)));

        /// <summary>
        /// Gets the calculated brush to use to color the item.
        /// </summary>
        public Brush ActualItemBrush
        {
            get
            {
                return (Brush)GetValue(ActualItemBrushProperty);
            }
            private set
            {
                SetValue(ActualItemBrushProperty, value);
            }
        }

        /// <summary>
        /// Identifies the ActualItemOutline dependency property.
        /// </summary>
        public static readonly DependencyProperty ActualItemOutlineProperty =
            DependencyProperty.Register(
            "ActualItemOutline",
            typeof(Brush),
            typeof(CustomDataContext),
            new PropertyMetadata(null,
                (o, e) => (o as CustomDataContext).RaisePropertyChanged("ActualItemOutline", e.OldValue, e.NewValue)));

        /// <summary>
        /// Gets the calculated brush to use to color the outline of the item.
        /// </summary>
        public Brush ActualItemOutline
        {
            get
            {
                return (Brush)GetValue(ActualItemOutlineProperty);
            }
            private set
            {
                SetValue(ActualItemOutlineProperty, value);
            }
        }

        #region ItemLabel dependency property

        internal const string ItemLabelPropertyName = "ItemLabel";

        /// <summary>
        /// Identifies the ItemLabel dependency property.
        /// </summary>
        public static readonly DependencyProperty ItemLabelProperty = DependencyProperty.Register(ItemLabelPropertyName, typeof(string), typeof(CustomDataContext),
            new PropertyMetadata(string.Empty, (sender, e) =>
            {
                (sender as CustomDataContext).RaisePropertyChanged(ItemLabelPropertyName, e.OldValue, e.NewValue);
            }
            ));

        /// <summary>
        /// Gets or sets the item label for the current data context.
        /// </summary>
        public string ItemLabel
        {
            get
            {
                return (string)GetValue(ItemLabelProperty);
            }
            set
            {
                SetValue(ItemLabelProperty, value);
            }
        }

        #endregion ItemLabel dependency property

        #region INotifyPropertyChanged Members

        /// <summary>
        /// Raises PropertyChanged event for a specified property name.
        /// </summary>
        /// <param name="propertyName">Property name</param>
        /// <param name="oldValue">Old property value</param>
        /// <param name="newValue">New property value</param>
        protected virtual void RaisePropertyChanged(string propertyName, object oldValue, object newValue)
        {
            switch (propertyName)
            {
                case "Series":
                    this.BindActualItemBrush(ItemBrush, ItemBrush);
                    this.BindActualItemOutline(ItemOutline, ItemOutline);
                    break;
                case "ItemBrush":
                    this.BindActualItemBrush(oldValue, newValue);
                    break;
                case "ItemOutline":
                    this.BindActualItemOutline(oldValue, newValue);
                    break;
            }
            if (PropertyChanged != null)
            {
                PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
            }
        }

        /// <summary>
        /// Raised when a property value changes.
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        #endregion INotifyPropertyChanged Members

        private void BindActualItemBrush(object oldValue, object newValue)
        {
            if (this.ItemBrush == null)
            {
                if (this.Series is MarkerSeries ||
                    (this.Series is Series && ((Series)this.Series).HasMarkers))
                {
                    BindingOperations.SetBinding(
                              this,
                              ActualItemBrushProperty,
                              new Binding("ActualMarkerBrush") { Source = this.Series });
                }
            }
            else if (oldValue == null && newValue != null)
            {
                BindingOperations.SetBinding(
                           this,
                           ActualItemBrushProperty,
                           new Binding("ItemBrush") { Source = this });
            }
        }

        private void BindActualItemOutline(object oldValue, object newValue)
        {
            if (this.ItemOutline == null)
            {
                if (this.Series is MarkerSeries ||
                    (this.Series is Series && ((Series)this.Series).HasMarkers))
                {
                    BindingOperations.SetBinding(
                              this,
                              ActualItemOutlineProperty,
                              new Binding("ActualMarkerOutline") { Source = this.Series });
                }
            }
            else if (oldValue == null && newValue != null)
            {
                BindingOperations.SetBinding(
                           this,
                           ActualItemOutlineProperty,
                           new Binding("ItemOutline") { Source = this });
            }
        }
    }


}
